/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   mp_earliestdeadline.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   September 14, 2005
 *
 *  Function        :   Earliest Deadline First scheduler for multi-processor
 *
 *  History         :
 *      14-09-05    :   Initial version.
 *
 * $Id: mp_earliestdeadline.cc,v 1.1.1.1 2007/10/02 10:59:47 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include "mp_earliestdeadline.h"

/**
 * MPearliestDeadlineScheduler ()
 * Constructor.
 */
MPearliestDeadlineScheduler::MPearliestDeadlineScheduler() : scheduler()
{
    schedule = NULL;
}
    
/**
 * ~MPearliestDeadlineScheduler ()
 * Desctructor
 */
MPearliestDeadlineScheduler::~MPearliestDeadlineScheduler()
{
}
    
/**
 * run ()
 * Run the scheduler on graph g.
 */
int MPearliestDeadlineScheduler::run(TimedSDFgraph *g)
{
    SPearliestDeadlineScheduler spScheduler;
    
    // Initialize
    schedule = new MPschedule();

    // Create a timed SDF graph for each cluster and schedule it
    for (SDFactorClustersIter iter = g->actorClustersBegin();
            iter != g->actorClustersEnd(); iter++)
    {
        SDFactorCluster *c = *iter;
    
        // Create graph
        TimedSDFgraph *gC = createClusterGraph(g, c);
        
        // Schedule graph on a single processor
        spScheduler.run(gC);

        // Add schedule to multi-processor schedule
        schedule->addSPschedule(spScheduler.getSchedule());
    }
    
    // Done
    return SCHEDULER_OK;
}
    
/**
 * check ()
 * Check that all conditions are met to schedule the graph g.
 *
 * Conditions:
 * - Each actor must belong to exactly one cluster
 */
int MPearliestDeadlineScheduler::check(TimedSDFgraph *g)
{
    vector<bool> actorInCluster(g->nrActors(), false);
    
    // Check all clusters
    for (SDFactorClustersIter iter = g->actorClustersBegin();
            iter != g->actorClustersEnd(); iter++)
    {
        SDFactorCluster *c = *iter;
        
        // Check all actors in cluster
        for (SDFactorsIter iterA = c->actorsBegin();
                iterA != c->actorsEnd(); iterA++)
        {
            SDFactor *a = *iterA;
            
            // Actor already seen in other cluster?
            if (actorInCluster[a->getId()])
                return SCHEDULER_ERROR;
            
            actorInCluster[a->getId()] = true;
        }
    }
    
    // Some actors not in a cluster?
    for (uint i = 0; i < g->nrActors(); i++)
        if (!actorInCluster[i])
            return SCHEDULER_ERROR;

    // Conditions met
    return SCHEDULER_OK;
}
    
/**
 * reset ()
 * Reset the scheduler.
 */
void MPearliestDeadlineScheduler::reset()
{
    delete schedule;
    schedule = NULL;
}

/**
 * createClusterGraph ()
 * The function creates a timed SDF graph which contains all actors (and
 * channels) that belong to the cluster. Dependencies with actors outside
 * the cluster are represented through dummy actors (with execution time 0).
 */
TimedSDFgraph *MPearliestDeadlineScheduler::createClusterGraph(TimedSDFgraph *g,
        SDFactorCluster *c)
{
    // Clone graph
    SDFcomponent component = SDFcomponent(g->getParent(), c->getId());
    TimedSDFgraph *gC = g->clone(component);
    
    // Remove all actors that do not belong to the cluster and/or do not have
    // a channel to an actor in the cluster
    for (SDFactorsIter iter = gC->actorsBegin();
            iter != gC->actorsEnd(); iter++)
    {
        TimedSDFactor *a = (TimedSDFactor*)(*iter);
        
        // Actor in the cluster?
        if (isActorInCluster(a,c))
            continue;
        
        // Actor connected to the cluster?
        if (isActorConnectedToCluster(a,c))
        {
            TimedSDFactor::Processor *p;
            
            // Actor becomes dummpy actor
            p = a->addProcessor("dummy");
            p->execTime = 0;
            a->setDefaultProcessor("dummy");
        }
        else
        {
            // Remove all channels the actor is connected to (including ports on
            // the actor at the other side)
            for (SDFportsIter iterP = a->portsBegin();
                    iterP != a->portsEnd(); iterP++)
            {
                SDFport *p = *iterP;
                SDFchannel *ch = p->getChannel();

                // Remove port on other side
                SDFport *po = ch->oppositePort(p);
                po->getActor()->removePort(po->getName());
                
                // Remove channel
                gC->removeChannel(ch->getName());                
            }
            
            // Remove the actor from the graph
            gC->removeActor(a->getName());
        }
    }
    
    return gC;
}

/**
 * isActorInCluster ()
 * The function returns true if the actor 'a' is in the cluster 'c', else
 * it returns false.
 */
bool MPearliestDeadlineScheduler::isActorInCluster(SDFactor *a,
        SDFactorCluster *c)
{
    for (SDFactorsIter iter = c->actorsBegin(); iter != c->actorsEnd(); iter++)
    {
        SDFactor *n = *iter;
        
        if (a->getId() == n->getId())
            return true;
    }
    
    return false;
}

/**
 * isActorConnectedToCluster ()
 * The function returns true if the actor is connected to an actor
 * that is part of the cluster, else it returns false. It is not checked
 * wether the actor itself is part of the cluster.
 */
bool MPearliestDeadlineScheduler::isActorConnectedToCluster(SDFactor *a,
        SDFactorCluster *c)
{
    for (SDFportsIter iter = a->portsBegin(); iter != a->portsEnd(); iter++)
    {
        SDFport *p = *iter;
        SDFchannel *ch = p->getChannel();
        SDFport *po = ch->oppositePort(p);
        
        if (isActorInCluster(po->getActor(), c))
            return true;
    }
    
    return false;
}
