/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   schedule.h
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   July 27, 2005
 *
 *  Function        :   Schedule for SDF graphs
 *
 *  History         :
 *      27-07-05    :   Initial version.
 *
 * $Id: schedule.h,v 1.1.1.1 2007/10/02 10:59:47 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#ifndef SDF_RESOURCE_ALLOCATION_SCHEDULING_SCHEDULE_H_INCLUDED
#define SDF_RESOURCE_ALLOCATION_SCHEDULING_SCHEDULE_H_INCLUDED

#include "../../base/timed/graph.h"

/**
 * Static-order schedule
 * A static-order schedule provides an ordering of actor firings for actors
 * mapped to the same processor. The schedule is described with a finite
 * state machine implemented through a vector with as initial state the zero-th
 * element of the vector.
 **/

typedef struct _Schedule
{
    CId id;
    SDFactor *actor;
    bool last;
    struct _Schedule *next;
    struct _Schedule *previous;
} Schedule;

// Foreward class definition
class SPschedule;

typedef list<SPschedule*>       SPschedules;
typedef SPschedules::iterator   SPschedulesIter;

/**
 * SPschedule
 * Container for single-processor schedule. A single-processor schedule
 * is a sequence of actor firings possibly written in a looped structure
 * (the loop body is a sub-schedule of the total schedule). For example,
 * the actor firing sequence 'ABBABCCABCC' can be expressed as
 * (1A)(2B)(2(1A)(1B)(2C)). So, at the top-level the schedule is the sequence
 * of (1A), (2B), (2(1A)(1B)(2C)).
 *
 * Infinite repetition of a sub-schedule is indicated by a repetition count
 * ('repCnt') of UINT_MAX.
 */
class SPschedule
{
public:
    // Constructor
    SPschedule(TimedSDFgraph *g, const CId id = 0)
        { repCnt = 1; scheduleId = id; sdfGraph = g;};
    
    // Destructor
    ~SPschedule();

    // Id
    CId getId() { return scheduleId; };

    // Repetition count
    uint repCnt;
    
    // Actor
    CId actorId;
    
    // Nested schedule
    SPschedules subSchedule;
    
    // Schedule
    void addActor(const CId actorId, const uint repCnt = 1);
    bool isSubSchedule() const { return (subSchedule.empty() ? false:true); };
    
    // Iterator
    SPschedulesIter subScheduleBegin() { return subSchedule.begin(); };
    SPschedulesIter subScheduleEnd() { return subSchedule.end(); };
    
    // Clear
    void clear() { subSchedule.clear(); repCnt = 1; };
    
    // Print
    void print(ostream &out);
    
private:
    CId scheduleId;
    TimedSDFgraph *sdfGraph;
};

/**
 * MPschedule
 * Container for multi-processor schedule. A multi-processor schedule
 * is a list of single-processor schedules.
 */
class MPschedule
{
public:
    // Constructor
    MPschedule() {};
    
    // Desctructor
    ~MPschedule();
    
    // Schedules
    SPschedule *getSPschedule(const CId id);
    SPschedule *newSPschedule(TimedSDFgraph *g, const CId id);
    void addSPschedule(SPschedule *s);
    void removeSPschedule(const CId id);
    uint nrSpSchedules() const { return spSchedules.size(); };
    
    // Iterators
    SPschedulesIter schedulesBegin() { return spSchedules.begin(); };
    SPschedulesIter schedulesEnd() { return spSchedules.end(); };
    
    // Print
    void print(ostream &out);
    
private:
    SPschedules spSchedules;    
};

#endif
